
from __future__ import division

#!/usr/bin/env python
#Clasifiers for prioritizing sRNAs
__author__ = "Erik Eppenhof"


import time
import os

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import functools
from itertools import cycle


from pylab import plot, show, savefig, xlim, figure, hold, ylim, legend, boxplot, setp, axes

import numpy as np
import pandas as pd
import seaborn as sns
from pandas import DataFrame
from sklearn.cross_validation import train_test_split
from sklearn.metrics import classification_report, accuracy_score, log_loss, recall_score, precision_score, average_precision_score, precision_recall_curve, auc, roc_curve
from sklearn.metrics import precision_recall_curve
from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.ensemble.weight_boosting import AdaBoostClassifier
from sklearn.model_selection import StratifiedKFold
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import LeaveOneOut
from scipy import interp
from runstats import Statistics
stats = Statistics()

#check version print(sklearn.__version__)

def load_data(data_set, testdata=False):
    '''
    This function loads the data from the relevant data files.
    '''

    #Filename training set and training set labels for R Capsulatus
    r_cap = 'training_R_capsulatus.csv'
    r_cap_labels = 'training_R_capsulatus_labels.csv'

    #Filename training set and training set labels for S Pyogenes
    s_pyo = 'training_S_pyogenes.csv'
    s_pyo_labels = 'training_S_pyogenes_labels.csv'
    
    #Filename training set and training set labels for S Enterica
    s_ent = 'training_S_enterica.csv'
    s_ent_labels = 'training_S_enterica_labels.csv'
    
    #Filename training set and training set labels for data of R Capsulatus, S Pyogenes and E Coli
    all_data_current = 'training_combined.csv'
    all_data_current_labels = 'training_combined_labels.csv'
   
    #Filename test set and test set labels for R Capsulatus
    rcap_validation = 'R_capsulatus_validation.csv'
    rcap_validation_labels = 'R_capsulatus_validation_labels.csv'

    #Filename test set and test set labels for S Pyogenes
    spyo_validation = 'S_pyogenes_validation.csv'
    spyo_validation_labels = 'S_pyogenes_validation_labels.csv'

    #Filename test set and test set labels for E Coli
    ecol_validation = 'E_coli_validation.csv'
    ecol_validation_labels = 'E_coli_validation_labels.csv'

    #Filename test set and test set labels for S Enterica
    sent_validation = 'S_enterica_validation.csv'
    sent_validation_labels = 'S_enterica_validation_labels.csv'

    #Filename test set and test set labels for M Tuberculosis
    mtub_validation = 'M_tuberculosis_validation.csv'
    mtub_validation_labels = 'M_tuberculosis_validation_labels.csv'

    run_on_server = False
    if(run_on_server == False):

        try:
            # Get data using path
            data_dir=os.path.dirname(os.path.abspath('')) + '/data/'
            print(data_dir)

            if (data_set == 'rcap'):
                train = pd.read_csv(data_dir + r_cap)
                trainlabels = pd.read_csv(data_dir + r_cap_labels)
            elif (data_set =='spyo'):
                train = pd.read_csv(data_dir + s_pyo)
                trainlabels = pd.read_csv(data_dir + s_pyo_labels)
            elif (data_set =='sent'):
                train = pd.read_csv(data_dir + s_ent)
                trainlabels = pd.read_csv(data_dir + s_ent_labels)
            elif (data_set =='all_data'):
                train = pd.read_csv(data_dir + all_data_current)
                trainlabels = pd.read_csv(data_dir + all_data_current_labels)
            elif(data_set == 'rcap_validation'):
                train = pd.read_csv(data_dir + rcap_validation)
                trainlabels = pd.read_csv(data_dir + rcap_validation_labels)
            elif (data_set == 'spyo_validation'):
                train = pd.read_csv(data_dir + spyo_validation)
                trainlabels = pd.read_csv(data_dir + spyo_validation_labels)
            elif (data_set == 'sent_validation'):
                train = pd.read_csv(data_dir + sent_validation)
                trainlabels = pd.read_csv(data_dir + sent_validation_labels)
            elif (data_set == 'mtub_validation'):
                train = pd.read_csv(data_dir + mtub_validation)
                trainlabels = pd.read_csv(data_dir + mtub_validation_labels)
            elif (data_set == 'ecol_validation'):
                train = pd.read_csv(data_dir + ecol_validation)
                trainlabels = pd.read_csv(data_dir + ecol_validation_labels)

        except IOError:
            # If the Unix path fails this makes it possible to load using Windows path
            print("error occurred with path")
            data_dir='..\\data\\'
            
     
        # Load the labels
        numerical_labels = ['SS', 'Pos10wrtsRNAStart', 'DistTerm', 'Distance', 'sameStrand', 'DownDistance', 'sameDownStrand']
        class_labels = ['Class']
        #Processing the labels of the files into the train and test sets:
        X_train=train[numerical_labels]

        trainset = np.column_stack((X_train,trainlabels['Class']))

        X_train = trainset[:,0:-1]
        X_valid = trainset[:,-1]

        #Based on whether train or test data is requested we return the correct one. Also the size of the data is printed.
        if testdata:
            print("Testing set has {0[0]} rows and {0[1]} columns/features".format(Xtrain.shape))
            #return (Xtest, test['id'])
        else:
            #print("Training set has {0[0]} rows and {0[1]} columns/features".format(Xtrain.shape))
            print("Dataset size:", len(X_train), len(X_valid))
            return(X_train.astype(float), X_valid.astype(float)) 
                #Y_train.astype(str), Y_valid.astype(str))

    elif(run_on_server==True):
        print("run_on_server equals True")
        #run other script: classifiers
        

def trainclf(data_set, validation_data, classifier_choice):
    X_train, X_valid = load_data(data_set, testdata = False)


    labels= ['SS','Pos10wrtsRNAStart','DistTerm','Distance','sameStrand','DownDistance','sameDownStrand']


    loo = LeaveOneOut()
    o = loo.get_n_splits(X_train)

    
    #---------CLASSIFIERS--------------------------------------------------------------------------
    
    # Logistic Regression
    lrg = LogisticRegression(class_weight='balanced')

    # Multilayer Perceptron
    mlp = MLPClassifier(hidden_layer_sizes = (400,), activation = 'logistic', solver = 'lbfgs',
                                                 max_iter = 200, alpha = 0.0001, verbose=0, learning_rate='constant', learning_rate_init=0.9, power_t=0.8)
    
    #AdaBoost with RF, random_state omitted
    ada = AdaBoostClassifier(RandomForestClassifier(n_jobs=3, n_estimators=100, max_depth=1))

    # Gradient boosting
    gbm=GradientBoostingClassifier(n_estimators=50, max_depth=15, max_features=7, min_samples_leaf=5,verbose=1, subsample=0.90)

    # Random Forest
    rfc = RandomForestClassifier(n_jobs=3, n_estimators=400, max_features = 2)
    #n_estimators of 400 is enough, it does not improve with more


    #choice of classifier
    classifier = ""
    if classifier_choice == "RF":
        classifier = rfc
    elif classifier_choice == "LR":
        classifier = lrg
    elif classifier_choice == "MP":
        classifier = mlp
    elif classifier_choice == "AB":
        classifier = ada
    elif classifier_choice == "GB":
        classifier = gbm
    else:
        raise Exception("Unknown classifier")

    #line width
    lw = 2
    
    i=0

    probabilities = np.asarray([])
    run_stat = []
    data_pos = []
    data_neg = []

    #Choice of testset in load_data:
    X_test_predict, X_valid_predict = load_data(validation_data, testdata = False)
    probas = classifier.fit(X_train, X_valid).predict_proba(X_test_predict)

    probabilities = np.append(probabilities,probas[:,1])
    prec, rec, thresholds = precision_recall_curve(X_valid_predict, probabilities)
    
    plt.figure(1)
    plt.plot(rec,prec)
    plt.savefig('prec_rec.png')
    
    rec_rev = list(reversed(rec))
    prec_rev = list(reversed(prec))

    value_at_set_rec_1 = np.interp(0.1,rec_rev,prec_rev)
    value_at_set_rec_2 = np.interp(0.2,rec_rev,prec_rev)
    value_at_set_rec_3 = np.interp(0.3,rec_rev,prec_rev)
    value_at_set_rec_4 = np.interp(0.4,rec_rev,prec_rev)
    value_at_set_rec_5 = np.interp(0.5,rec_rev,prec_rev)
    value_at_set_rec_6 = np.interp(0.6,rec_rev,prec_rev)
    value_at_set_rec_7 = np.interp(0.7,rec_rev,prec_rev)
    value_at_set_rec_8 = np.interp(0.8,rec_rev,prec_rev)
    value_at_set_rec_9 = np.interp(0.9,rec_rev,prec_rev)
    value_at_set_rec_10 = np.interp(1.0,rec_rev,prec_rev)
    
    f1_p10 = (2*(0.10*value_at_set_rec_1))/(0.10+value_at_set_rec_1)
    f1_p40 = (2*(0.40*value_at_set_rec_4))/(0.40+value_at_set_rec_4)
    f1_p60 = (2*(0.60*value_at_set_rec_6))/(0.60+value_at_set_rec_6)
    f1_p90 = (2*(0.90*value_at_set_rec_9))/(0.90+value_at_set_rec_9)

    print(classifier_choice, "\t", data_set, "\t", validation_data, "\tP10\t",value_at_set_rec_1, "\tF1\t", f1_p10)
    print(classifier_choice, "\t",data_set, "\t", validation_data, "\tP40\t",value_at_set_rec_4, "\tF1\t", f1_p40)
    print(classifier_choice, "\t",data_set, "\t", validation_data, "\tP60\t",value_at_set_rec_6, "\tF1\t", f1_p60)
    print(classifier_choice, "\t", data_set, "\t", validation_data, "\tP90\t", value_at_set_rec_9, "\tF1\t", f1_p90)

    plt.figure(2)
    fpr, tpr, thresholds = roc_curve(X_valid_predict, probabilities)
    plt.plot([0,1],[0,1], linestyle='--', lw=lw, color='k', label = 'Luck')

    plt.plot(fpr, tpr, color='g', linestyle='--', label = 'Mean ROC (area = %0.2f)' % auc(fpr,tpr), lw=lw)
    print(classifier_choice, "\t",data_set, "\t", validation_data, "\tAUC\t",auc(fpr,tpr))
    print(classifier_choice, "\t",data_set, "\t", validation_data, "\tAUPRC\t", auc(rec,prec))

    plt.xlim([-0.05, 1.05])
    plt.ylim([-0.05, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('ROC')
    plt.legend(loc = 'lower right')
    if(data_set == 'rcap'):
        plt.savefig('roc_rcap.png')
    if(data_set == 'spyo'):
        plt.savefig('roc_spyo.png')
    if(data_set == 'ecol'):
        plt.savefig('roc_ecol.png')
    if(data_set == 'all_data'):
        plt.savefig('roc_all_data.png')
    if(data_set == 'sent'):
        plt.savefig('roc_sent.png')

    return data_pos, data_neg

def main():
    print(" - Start.")
    start_time = time.time()
    #Run 5 runs of training
    for i in range(0,5):
        print("training run ", i)
        #Run for all classifiers, training data and validation data
        for classifier in ["LR", "MP", "AB", "GB", "RF"]:
            for ds in ['all_data', 'spyo', 'rcap', 'sent']:
                for vd in ['rcap_validation','spyo_validation', 'sent_validation','ecol_validation','mtub_validation']:
                    data_pos, data_neg = trainclf(data_set = ds, validation_data = vd, classifier_choice =classifier)
    print(" - Finished.")
    print("--- execution took %s seconds ---" % (time.time() - start_time))
    quit()

main()
